# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import grpc
from kubemq.basic import configuration_loader
from kubemq.grpc import kubemq_pb2_grpc


class ServerAddressNotSuppliedException(Exception):
    def __init__(self):
        self.message = "Server Address was not supplied"

    def __str__(self):
        return str(self.message)


class GrpcClient:
    _kubemq_address = None
    _metadata = None

    _channel = None
    _client = None

    def __init__(self, encryptionHeader):
        self._init_registration(encryptionHeader)

    def get_kubemq_client(self):
        if not self._client:
            if not self._channel:
                kubemq_address = self.get_kubemq_address()
                client_cert_file = configuration_loader.get_certificate_file()
                if client_cert_file:
                    # Open SSL/TLS connection
                    with open(client_cert_file, 'rb') as f:
                        credentials = grpc.ssl_channel_credentials(f.read())
                    self._channel = grpc.secure_channel(kubemq_address, credentials)
                else:
                    # Open Insecure connection
                    self._channel = grpc.insecure_channel(kubemq_address)
            self._client = kubemq_pb2_grpc.kubemqStub(self._channel)

        return self._client

    def get_kubemq_address(self):
        if self._kubemq_address:
            return self._kubemq_address

        self._kubemq_address = configuration_loader.get_server_address()

        if not self._kubemq_address:
            raise ServerAddressNotSuppliedException()

        return self._kubemq_address

    def _init_registration(self, encryptionHeader):
        if encryptionHeader:
            self._metadata = [("authorization", encryptionHeader)]



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from kubemq.tools.id_generator import get_next_id
from kubemq.grpc import Request as InnerRequest


class Request:
    """Represents the Request used in RequestReply to send information using the KubeMQ."""

    def __init__(self, inner_request=None, request_id=None, request_type=None, client_id=None, channel=None,
                 reply_channel=None, metadata=None, body=None, timeout=None, cache_key=None, cache_ttl=None, tags=None):
        """Initializes a new instance of the Request with a set of parameters."""
        if inner_request:
            self.request_id = inner_request.RequestID or get_next_id()
            """Represents a Request identifier."""

            self.request_type = inner_request.RequestTypeData
            """Represents the type of the request operation."""

            self.client_id = inner_request.ClientID
            """Represents the sender ID that the Request will be send under."""

            self.channel = inner_request.Channel
            """Represents The channel name to send to using the KubeMQ."""

            self.reply_channel = inner_request.ReplyChannel
            """Represents The channel name to return response to."""

            self.metadata = inner_request.Metadata
            """Represents metadata text attached to the request."""

            self.body = inner_request.Body
            """Represents The content of the request."""

            self.timeout = inner_request.Timeout
            """Represents the limit for waiting for response (Milliseconds)."""

            self.cache_key = inner_request.CacheKey
            """Represents if the request should be saved from Cache and under what "Key"(Str) to save it."""

            self.cache_ttl = inner_request.CacheTTL
            """Cache time to live : for how long does the request should be saved in Cache."""

            self.tags = inner_request.Tags
            """Represents key value pairs that help distinguish the message"""

        else:
            self.request_id = request_id
            self.request_type = request_type
            self.client_id = client_id
            self.channel = channel
            self.reply_channel = reply_channel
            self.metadata = metadata
            self.body = body
            self.timeout = timeout
            self.cache_key = cache_key
            self.cache_ttl = cache_ttl
            self.tags = tags

    def convert(self):
        """Convert a Request to an InnerRequest"""
        return InnerRequest(
            RequestID=self.request_id or get_next_id(),
            RequestTypeData=self.request_type.value,
            ClientID=self.client_id or "",
            Channel=self.channel,
            Metadata=self.metadata or "",
            Body=self.body,
            Timeout=self.timeout,
            CacheKey=self.cache_key,
            CacheTTL=self.cache_ttl,
            Tags=self.tags
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class AckAllMessagesResponse:
    def __init__(self, ack_all_messages_response=None):
        if ack_all_messages_response:
            self.request_id = ack_all_messages_response.RequestID
            """Represents Unique identifier for the Request."""

            self.is_error = ack_all_messages_response.IsError
            """Returned from KubeMQ, false if no error."""

            self.error = ack_all_messages_response.Error
            """Error message, valid only if IsError true."""
            if ack_all_messages_response.AffectedMessages:
                self.affected_messages = ack_all_messages_response.AffectedMessages
            else:
                self.affected_messages = None
            """"Number of affected messages."""

    def __repr__(self):
        return "<AckAllMessagesResponse request_id:%s is_error:%s error:%s affected_messages:%s>" % (
            self.request_id,
            self.is_error,
            self.error,
            self.affected_messages
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import enum
import logging
import threading
from queue import Queue

from grpc import StatusCode

from kubemq.queue.message import Message
from kubemq.basic.grpc_client import GrpcClient
from kubemq.queue.transaction_messages import TransactionMessagesResponse
from kubemq.queue.transaction_messages import create_stream_queue_message_ack_request
from kubemq.queue.transaction_messages import create_stream_queue_message_extend_visibility_request
from kubemq.queue.transaction_messages import create_stream_queue_message_modify_request
from kubemq.queue.transaction_messages import create_stream_queue_message_receive_request
from kubemq.queue.transaction_messages import create_stream_queue_message_reject_request
from kubemq.queue.transaction_messages import create_stream_queue_message_resend_request
from kubemq.tools.id_generator import get_next_id

logger = logging.getLogger(__name__)


class ControlMessages(enum.Enum):
    CLOSE_STREAM = enum.auto()


class Transaction(GrpcClient):
    """Represents a MessageQueue pattern.

    Attributes:
        queue: should be called from queue.transaction()".
        stream: If stream is active.
        inner_stream: represent the stream receive from kubemq
        _kubemq_address:represent kubemq connection string.
        lock:inner lock for checking stream
    """

    def __init__(self, queue):
        """
        Initializes a new Transaction using MessageQueue .
        :param queue: should be called from queue.transaction()".
        """
        super().__init__(queue.encryptionHeader)
        self.queue = queue
        self.stream = False
        self.inner_stream = None
        self._kubemq_address = queue.get_kubemq_address()
        self.client = self.get_kubemq_client()
        self.lock = threading.Lock()
        self.stream_observer = None

    def receive(self, visibility_seconds=1, wait_time_seconds=1):
        """Receive queue messages request , waiting for response or timeout."""
        if self.open_stream():
            raise Exception("Stream already open , please call ack")

        try:
            msg = create_stream_queue_message_receive_request(self.queue, visibility_seconds, wait_time_seconds)
            self.stream_observer.put(msg)

            return TransactionMessagesResponse(next(self.inner_stream))
        except Exception as e:
            logger.exception("Exception in receive: '%s'" % e)
            raise

    def ack_message(self, msg_sequence):
        """Will mark Message de-queued on queue."""
        if not self.check_call_is_in_transaction():
            raise Exception("no active message to ack, call Receive first")

        try:
            msg = create_stream_queue_message_ack_request(self.queue, msg_sequence)
            self.stream_observer.put(msg)

            return TransactionMessagesResponse(next(self.inner_stream))
        except Exception as e:
            logger.exception("Exception in ack:'%s'" % e)
            raise

    def rejected_message(self, msg_sequence):
        """Will return message to queue."""
        if not self.check_call_is_in_transaction():
            return TransactionMessagesResponse(None, None, True, "no active message to reject, call Receive first")
        else:
            try:
                msg = create_stream_queue_message_reject_request(self.queue, msg_sequence)
                self.stream_observer.put(msg)

                return TransactionMessagesResponse(next(self.inner_stream))
            except Exception as e:
                logger.exception("Exception in reject:'%s'" % e)
                raise

    def extend_visibility(self, visibility_seconds):
        """Extend the visibility time for the current receive message."""
        if not self.check_call_is_in_transaction():
            return TransactionMessagesResponse(None, None, True,
                                               "no active message to extend visibility, call Receive first")
        else:
            try:
                msg = create_stream_queue_message_extend_visibility_request(self.queue, visibility_seconds)
                self.stream_observer.put(msg)

                return TransactionMessagesResponse(next(self.inner_stream))
            except Exception as e:
                logger.exception("Exception in Extend:'%s'" % e)
                raise

    def resend(self, queue_name):
        """Resend the current received message to a new channel and ack the current message."""
        if not self.check_call_is_in_transaction():
            return TransactionMessagesResponse(None, None, True, "no active message to resend, call Receive first")
        else:
            try:
                msg = create_stream_queue_message_resend_request(self.queue, queue_name)
                self.stream_observer.put(msg)

                return TransactionMessagesResponse(next(self.inner_stream))
            except Exception as e:
                logger.exception("Exception in resend:'%s'" % e)
                raise

    def modify(self, msg):
        """Resend the new message to a new channel."""
        if not self.check_call_is_in_transaction():
            return TransactionMessagesResponse(None, None, True,
                                               "no active message to modify, call Receive first")
        else:
            try:
                if isinstance(msg, Message):
                    msg.client_id = msg.client_id or self.queue.client_id
                    msg = msg.convert_to_queue_message()
                msg.ClientID = msg.ClientID
                msg.MessageID = get_next_id()

                msg = create_stream_queue_message_modify_request(self.queue, msg)
                self.stream_observer.put(msg)

                return TransactionMessagesResponse(next(self.inner_stream))
            except Exception as e:
                logger.exception("Exception in resend:'%s'" % e)
                raise

    def open_stream(self):
        with self.lock:
            if self.stream:
                return True

            stream_observer = Queue()

            def stream():
                while True:
                    msg = stream_observer.get()

                    if msg == ControlMessages.CLOSE_STREAM:
                        stream_observer.task_done()
                        break

                    if msg:
                        yield msg
                        stream_observer.task_done()

            def done(call):
                stream_observer.put(ControlMessages.CLOSE_STREAM)
                self.stream = False

            self.inner_stream = self.client.StreamQueueMessage(stream(), metadata=self.queue._metadata)
            self.inner_stream.add_done_callback(done)

            self.stream_observer = stream_observer
            self.stream = True
            return False

    def close_stream(self):
        with self.lock:
            if self.stream:
                self.stream = False
                self.inner_stream.cancel()
                return True
            else:
                logger.error("Stream is closed")
                return False

    def check_call_is_in_transaction(self):
        with self.lock:
            return self.stream



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from kubemq.queue import send_message_result as create_result
from kubemq.grpc import QueueMessagesBatchResponse as QueueMessagesBatchResponse


def convert_from_queue_messages_batch_response(queue_messages_batch_response):
    """Convert from QueueMessagesBatchResponse to SendBatchMessageResult"""
    batch_message_result = SendBatchMessageResult()
    batch_message_result.batch_id = queue_messages_batch_response.BatchID
    batch_message_result.result = queue_messages_batch_response.Results
    batch_message_result.have_errors = queue_messages_batch_response.HaveErrors
    return batch_message_result


class SendBatchMessageResult:
    def __init__(self, queue_messages_batch_response=None):
        if queue_messages_batch_response:
            self.batch_id = queue_messages_batch_response.BatchID
            """Represents Unique identifier for the Request."""

            self.have_errors = queue_messages_batch_response.HaveErrors
            """Returned from KubeMQ, false if no error."""

            self.result = queue_messages_batch_response.Error
            """Error message, valid only if IsError true."""

    def __repr__(self):
        return "<SendMessageResult batch_id:%s have_errors:%s result:%s>" % (
            self.batch_id,
            self.have_errors,
            self.result
        )

    def convert_to_send_message_result(self, results):
        """convert a few results to SendMessageResult """
        for result in results:
            yield create_result.SendMessageResult(result)



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import threading
import grpc
from kubemq.grpc import Empty
from kubemq.basic.grpc_client import GrpcClient
from kubemq.commandquery.request_receive import RequestReceive
from kubemq.tools.listener_cancellation_token import ListenerCancellationToken

logger = logging.getLogger(__name__)


class Responder(GrpcClient):
    """An instance that responsible on receiving request from the kubeMQ."""

    def __init__(self, kubemq_address=None, encryptionHeader=None):
        """
        Initialize a new Responder to subscribe to Response.

        :param str kubemq_address: KubeMQ server address. if None will be parsed from Config or environment parameter.
        """
        GrpcClient.__init__(self, encryptionHeader)
        if kubemq_address:
            self._kubemq_address = kubemq_address

    def ping(self):
        """ping check connection to the kubemq"""
        ping_result = self.get_kubemq_client().Ping(Empty())
        logger.debug("Responder KubeMQ address:%s ping result:%s'" % (self._kubemq_address, ping_result))
        return ping_result

    def subscribe_to_requests(self, subscribe_request, handler, error_handler,
                              listener_cancellation_token=ListenerCancellationToken()):
        """
        Register to kubeMQ Channel using handler.
        :param SubscribeRequest subscribe_request: represent by that will determine the subscription configuration.
        :param handler: Method the perform when receiving RequestReceive
        :param error_handler: Method the perform when receiving error from kubemq
        :param listener_cancellation_token: cancellation token, once cancel is called will cancel the subscribe to kubemq
        :return: A thread running the Subscribe Request.
        """

        if not subscribe_request.channel:
            raise ValueError("Channel parameter is mandatory.")

        if not subscribe_request.is_valid_type("CommandQuery"):
            raise ValueError("Invalid Subscribe Type for this Class.")

        inner_subscribe_request = subscribe_request.to_inner_subscribe_request()

        call = self.get_kubemq_client().SubscribeToRequests(inner_subscribe_request, metadata=self._metadata)

        def subscribe_task(listener_cancellation_token):
            while True:
                try:
                    event_receive = call.next()
                    logger.debug("Responder InnerRequest. ID:'%s', Channel:'%s', ReplyChannel:'%s tags:'%s''" % (
                        event_receive.RequestID,
                        event_receive.Channel,
                        event_receive.ReplyChannel,
                        event_receive.Tags
                    ))

                    if handler:
                        try:
                            response = handler(RequestReceive(event_receive))

                            logger.debug("Responder InnerResponse. ID:'%s', ReplyChannel:'%s'" % (
                                response.request_id,
                                response.reply_channel
                            ))

                            self.get_kubemq_client().SendResponse(response.convert(), None, self._metadata)
                        except grpc.RpcError as error:
                            if (listener_cancellation_token.is_cancelled):
                                logger.info("Sub closed by listener request")
                                error_handler(str(error))
                            else:
                                logger.exception("Subscriber Received Error: Error:'%s'" % (error))
                                error_handler(str(error))
                        except Exception as e:
                            logger.exception("Subscriber Received Error: Error:'%s'" % (e))
                            error_handler(str(e))

                except Exception as e:
                    logger.exception("An exception occurred while listening for request:'%s'" % (e))
                    error_handler(str(e))
                    raise  # re-raise the original exception, keeping full stack trace

        def check_sub_to_valid(listener_cancellation_token):
            while True:
                if (listener_cancellation_token.is_cancelled):
                    logger.info("Sub closed by listener request")
                    call.cancel()
                    return

        thread = threading.Thread(target=subscribe_task, args=(listener_cancellation_token,))
        thread.daemon = True
        thread.start()

        listener_thread = threading.Thread(target=check_sub_to_valid, args=(listener_cancellation_token,))
        listener_thread.daemon = True
        listener_thread.start()
        return thread



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from kubemq.queue.message import Message as Message


class ReceiveMessagesResponse:
    def __init__(self, receive_queue_messages_response=None):
        if receive_queue_messages_response:
            self.error = receive_queue_messages_response.Error
            self.is_error = receive_queue_messages_response.IsError
            self.is_peek = receive_queue_messages_response.IsPeak
            my_messages = []
            for message in receive_queue_messages_response.Messages:
                my_messages.append(Message(message))
            self.messages = receive_queue_messages_response.Messages
            self.messages_expired = receive_queue_messages_response.MessagesExpired
            self.messages_received = receive_queue_messages_response.MessagesReceived
            self.request_id = receive_queue_messages_response.RequestID

    def __repr__(self):
        return "<ReceiveMessagesResponse error:%s is_error:%s is_peek:%s messages:%s messages_expired:%s messages_received:%s request_id:%s>" % (
            self.error,
            self.is_error,
            self.is_peek,
            self.messages,
            self.messages_expired,
            self.messages_received,
            self.request_id,
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class EventReceive:
    def __init__(self, inner_event_receive=None):
        if inner_event_receive:
            self.event_id = inner_event_receive.EventID
            """Represents a event identifier."""

            self.channel = inner_event_receive.Channel
            """Represents The channel name to send to using the KubeMQ."""

            self.metadata = inner_event_receive.Metadata
            """Represents text as str"""

            self.body = inner_event_receive.Body
            """Represents The content of the KubeMQRequestReceive."""

            self.timestamp = inner_event_receive.Timestamp
            """Represents the timestamp the message arrived """

            self.sequence = inner_event_receive.Sequence
            """"Represents the sequnce of the event (set by Kubemq)"""

            self.tags = inner_event_receive.Tags
            """Represents key value pairs that help distinguish the message"""



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from kubemq.grpc import QueueMessage as QueueMessage
from kubemq.grpc import QueueMessagesBatchRequest
from kubemq.tools.id_generator import get_next_id as get_next_id


class Message:
    def __init__(self, queue_message=None, message_id=None, client_id=None, queue=None, metadata=None, body=None,
                 tags=None, attributes=None, policy=None):
        if queue_message:
            self.message_id = queue_message.MessageID
            """Represents identifier to help distinguish the message"""

            self.client_id = queue_message.ClientID
            """Represents identifier to help distinguish the sender"""

            self.queue = queue_message.Channel
            """Represents the channel name to send the message to."""

            self.metadata = queue_message.Metadata
            """Represents text as str."""

            self.body = queue_message.Body
            """Represents The content of the Message."""

            self.tags = tags
            """Represents key value pairs that help distinguish the message"""

            self.attributes = queue_message.Attributes
            """Contain general data on the message."""

            self.policy = queue_message.Policy
            """A set of 'rules' that can be assign to the message"""
        else:
            self.message_id = message_id
            """Represents identifier to help distinguish the message"""

            self.client_id = client_id
            """Represents identifier to help distinguish the sender"""

            self.queue = queue
            """Represents the channel name to send the message to."""

            self.metadata = metadata
            """Represents text as str."""

            self.body = body
            """Represents The content of the Message."""

            self.tags = tags
            """Represents key value pairs that help distinguish the message"""

            self.attributes = attributes
            """Contain general data on the message."""

            self.policy = policy
            """A set of 'rules' that can be assign to the message"""

    def __repr__(self):
        return "<Message message_id:%s client_id:%s queue:%s metadata:%s body:%s tags:%s attributes:%s policy:%s>" % (
            self.message_id,
            self.client_id,
            self.queue,
            self.metadata,
            self.body,
            self.tags,
            self.attributes,
            self.policy,
        )

    def convert_to_queue_message(self, queue=None):
        """Convert a Message to an QueueMessage"""
        return QueueMessage(
            MessageID=self.message_id or get_next_id(),
            ClientID=self.client_id or queue.client_id,
            Channel=self.queue or queue.queue_name,
            Metadata=self.metadata,
            Body=self.body,
            Tags=self.tags or "",
            Attributes=self.attributes,
            Policy=self.policy
        )


def convert_queue_message_batch_request(uuid, messages):
    """Convert a messages to an QueueMessagesBatchRequest"""
    return QueueMessagesBatchRequest(
        BatchID=uuid,
        Messages=messages,
    )


def to_queue_messages(messages, queue=None, channel_name=None):
    """Convert  messages to a single message"""
    queue_messages = []
    for message in messages:
        if message.queue is None and channel_name is not None:
            message.queue = channel_name
        queue_messages.append(message.convert_to_queue_message(queue))

    return queue_messages



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import threading

from kubemq.grpc import Event as kubeEvent

_lock = threading.Lock()
_counter = 0


def get_next_id():
    """Get an unique thread safety ID between 1 to 65535"""
    global _lock, _counter

    with _lock:
        if _counter == 65535:
            _counter = 1
        else:
            _counter += 1

        return str(_counter)


class Event:

    def __init__(self, channel=None, client_id=None, store=None, event_id=None, body=None, metadata=None, tags=None):
        self.channel = channel
        """Represents The channel name to send to using the KubeMQ ."""

        self.client_id = client_id
        """Represents the sender ID that the events will be send under."""

        self.store = store
        """Represents if the events should be send to persistence."""

        self.event_id = event_id
        """Represents a Event identifier."""

        self.body = body
        """Represents The content of the LowLevel.Event."""

        self.metadata = metadata
        """Represents text as str."""

        self.tags = tags
        """Represents key value pairs that help distinguish the message"""

    def from_inner_event(self, inner_event):
        self.channel = inner_event.Channel
        self.metadata = inner_event.Metadata
        self.body = inner_event.Body
        self.event_id = inner_event.EventID or get_next_id()
        self.client_id = inner_event.ClientID
        self.store = inner_event.Store
        self.tags = inner_event.Tags

    def to_inner_event(self):
        return kubeEvent(
            Channel=self.channel,
            Metadata=self.metadata or "",
            Body=self.body,
            EventID=self.event_id or get_next_id(),
            ClientID=self.client_id,
            Store=self.store,
            Tags=self.tags or ""
        )

    def __repr__(self):
        return "<LowLevel.Event channel:%s client_id:%s store:%s event_id:%s body:%s metadata:%s tags:%s>" % (
            self.channel,
            self.client_id,
            self.store,
            self.event_id,
            self.body,
            self.metadata,
            self.tags
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ChannelParameters:

    def __init__(self, channel_name=None, client_id=None, timeout=None, cache_key=None, cache_ttl=None,
                 request_type=None, kubemq_address=None, encryptionHeader=None):
        """
        Initializes a new instance of the ChannelParameters

        :param str channel_name: Represents The channel name to send to using the KubeMQ.
        :param str client_id: Represents the sender ID that the messages will be send under.
        :param int timeout: Represents the limit for waiting for response (Milliseconds).
        :param str cache_key: Represents if the request should be saved from Cache and under what "Key"(Str) to save it.
        :param int cache_ttl: Cache time to live : for how long does the request should be saved in Cache.
        :param RequestType request_type: Represents the type of request operation.
        :param str kubemq_address: Represents The address of the KubeMQ server.
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """

        self.channel_name = channel_name
        """Represents The channel name to send to using the KubeMQ."""

        self.client_id = client_id
        """Represents the sender ID that the messages will be send under."""

        self.timeout = timeout
        """Represents the limit for waiting for response (Milliseconds)."""

        self.cache_key = cache_key
        """Represents if the request should be saved from Cache and under what "Key"(str) to save it."""

        self.cache_ttl = cache_ttl
        """Cache time to live : for how long does the request should be saved in Cache."""

        self.request_type = request_type
        """Represents the type of request operation."""

        self.kubemq_address = kubemq_address
        """Represents The address of the KubeMQ server."""

        self.encryptionHeader = encryptionHeader
        """the header for authentication using kubemq."""



import logging
import threading
import grpc
from kubemq.grpc import Empty
from kubemq.basic.grpc_client import GrpcClient
from kubemq.events.event_receive import EventReceive
from kubemq.subscription import SubscribeType, EventsStoreType
from kubemq.tools.listener_cancellation_token import ListenerCancellationToken

logger = logging.getLogger(__name__)

class Subscriber(GrpcClient):

    def __init__(self, kubemq_address=None, encryptionHeader=None):
        """
        Initialize a new Sender under the requested KubeMQ Server Address.

        :param str kubemq_address: KubeMQ server address. if None will be parsed from Config or environment parameter.
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """
        GrpcClient.__init__(self, encryptionHeader)
        if kubemq_address:
            self._kubemq_address = kubemq_address

    def ping(self):
        """ping check connection to the kubemq"""
        ping_result = self.get_kubemq_client().Ping(Empty())
        logger.debug("event subscriber KubeMQ address:%s ping result:%s'" % (self._kubemq_address, ping_result))
        return ping_result

    def subscribe_to_events(self, subscribe_request, handler, error_handler,
                            listener_cancellation_token=ListenerCancellationToken()):
        """
        Register to kubeMQ Channel using handler.
        :param SubscribeRequest subscribe_request: represent by that will determine the subscription configuration.
        :param handler: Method the perform when receiving EventReceive
        :param error_handler: Method the perform when receiving error from kubemq
        :param listener_cancellation_token: cancellation token, once cancel is called will cancel the subscribe to kubemq
        :return: A thread running the Subscribe Request.
        """

        if not subscribe_request.channel:
            raise ValueError("channel parameter is mandatory.")

        if not subscribe_request.is_valid_type("Events"):
            raise ValueError("Invalid Subscribe Type for this Class.")

        if subscribe_request.subscribe_type == SubscribeType.EventsStore:
            if not subscribe_request.client_id:
                raise ValueError("client_id parameter is mandatory.")
            if subscribe_request.events_store_type == EventsStoreType.Undefined:
                raise ValueError("events_store_type parameter is mandatory.")

        inner_subscribe_request = subscribe_request.to_inner_subscribe_request()

        call = self.get_kubemq_client().SubscribeToEvents(inner_subscribe_request, metadata=self._metadata)

        def subscribe_to_event(listener_cancellation_token):
            try:
                while True:
                    event_receive = call.next()

                    logger.info("Subscriber Received Event: EventID:'%s', Channel:'%s', Body:'%s Tags:%s'" % (
                        event_receive.EventID,
                        event_receive.Channel,
                        event_receive.Body,
                        event_receive.Tags
                    ))

                    handler(EventReceive(event_receive))
            except grpc.RpcError as error:
                if (listener_cancellation_token.is_cancelled):
                    logger.info("Sub closed by listener request")
                    error_handler(str(error))
                else:
                    logger.exception("Subscriber Received Error: Error:'%s'" % (error))
                    error_handler(str(error))
            except Exception as e:
                logger.exception("Subscriber Received Error: Error:'%s'" % (e))
                error_handler(str(e))

        def check_sub_to_valid(listener_cancellation_token):
            while True:
                if (listener_cancellation_token.is_cancelled):
                    logger.info("Sub closed by listener request")
                    call.cancel()
                    return

        thread = threading.Thread(target=subscribe_to_event, args=(listener_cancellation_token,))
        thread.daemon = True
        thread.start()

        listener_thread = threading.Thread(target=check_sub_to_valid, args=(listener_cancellation_token,))
        listener_thread.daemon = True
        listener_thread.start()
        return thread



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ChannelParameters:
    """Configuration parameters for a Channel."""

    def __init__(self, channel_name=None, client_id=None, store=None, kubemq_address=None, return_result=False,
                 encryptionHeader=None):
        """
        Initializes a new instance of the ChannelParameters class with set parameters.

        :param channel_name: Represents The channel name to send to using the KubeMQ.
        :param client_id: Represents the sender ID that the messages will be send under.
        :param store: Represents the channel persistence property.
        :param kubemq_address: Represents The address of the KubeMQ server.
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """
        self.channel_name = channel_name
        self.client_id = client_id
        self.store = store
        self.kubemq_address = kubemq_address
        self.encryptionHeader = encryptionHeader
        """Represents the header for authentication using kubemq."""

    def __repr__(self):
        return "<ChannelParameters channel_name:%s client_id:%s store:%s kubemq_address:%s encryptionHeader:%s>" % (
            self.channel_name,
            self.client_id,
            self.store,
            self.kubemq_address,
            self.encryptionHeader
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys

from kubemq.commandquery.lowlevel.initiator import Initiator
from kubemq.commandquery.lowlevel.request import Request as LowLevelRequest


class Channel:
    """Represents a Initiator with predefined parameters."""

    def __init__(self, channel_parameters=None, request_type=None, channel_name=None, client_id=None, timeout=None,
                 cache_key=None,
                 cache_ttl=None, kubemq_address=None, encryptionHeader=None):
        # Initializes a new instance of the RequestChannel class using RequestChannelParameters.
        if channel_parameters:
            self.request_type = channel_parameters.request_type
            self.channel_name = channel_parameters.channel_name
            self.client_id = channel_parameters.client_id
            self.timeout = channel_parameters.timeout
            self.cache_key = channel_parameters.cache_key
            self.cache_ttl = channel_parameters.cache_ttl
            self.kubemq_address = channel_parameters.kubemq_address
            self.encryptionHeader = channel_parameters.encryptionHeader
        # Initializes a new instance of the RequestChannel class using a set of parameters.
        else:
            self.request_type = request_type

            self.channel_name = channel_name
            """Represents The channel name to send to using the KubeMQ."""

            self.client_id = client_id
            """Represents the sender ID that the Request will be send under."""

            self.timeout = timeout
            """Represents the limit for waiting for response (Milliseconds)."""

            self.cache_key = cache_key
            """Represents if the request should be saved from Cache and under what "Key"(System.String) to save it."""

            self.cache_ttl = cache_ttl
            """Cache time to live : for how long does the request should be saved in Cache."""

            self.kubemq_address = kubemq_address

            self.encryptionHeader = encryptionHeader
            """the header for authentication using kubemq."""

        # Validate arguments

        if not self.channel_name:
            raise ValueError("channel_name argument is mandatory")

        if not self.request_type:
            raise ValueError("request_type argument is mandatory")

        if not self.timeout and timeout <= 0:
            raise ValueError("timeout argument is mandatory and must between 1 to {}" % sys.maxsize)

        self._initiator = Initiator(self.kubemq_address, self.encryptionHeader)

    def send_request(self, request, override_params=None):
        """
        Publish a single request using the KubeMQ, response will return in the passed handler.

        Parameters
        ----------
        :param Request request: The request that will be sent to the kubeMQ.
        :param RequestParameters override_params: overriding `timeout`, `cache-key` and `cache_ttl` for the Request.
        :return Response
        """
        return self._initiator.send_request(self.create_low_level_request(request, override_params))

    def send_request_async(self, request, handler, override_params=None):
        """
        Publish a single async request using the KubeMQ, response will return in the passed handler.

        Parameters
        ----------
        :param Request request: The request that will be sent to the kubeMQ.
        :param handler: handler for processing response
        :param RequestParameters override_params: overriding `timeout`, `cache-key` and `cache_ttl` for the Request.
        :return Response
        """
        return self._initiator.send_request_async(self.create_low_level_request(request, override_params), handler)

    def create_low_level_request(self, request, override_params=None):
        request = LowLevelRequest(
            request_type=self.request_type,
            channel=self.channel_name,
            client_id=self.client_id,
            timeout=self.timeout,
            cache_key=self.cache_key,
            cache_ttl=self.cache_ttl,
            request_id=request.request_id,
            body=request.body,
            metadata=request.metadata,
            tags=request.tags
        )

        if override_params:
            if override_params.timeout:
                request.timeout = override_params.timeout
            if override_params.cache_key:
                request.cache_key = override_params.cache_key
            if override_params.cache_ttl:
                request.cache_ttl = override_params.cache_ttl

        return request



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from kubemq.grpc import Empty
from kubemq.basic.grpc_client import GrpcClient
from kubemq.events.lowlevel.event import Event
from kubemq.events.result import Result

logger = logging.getLogger(__name__)


class Sender(GrpcClient):
    """Represents the instance that is responsible to send events to the kubemq."""

    def __init__(self, kubemq_address=None, encryptionHeader=None):
        """
        Initialize a new Sender under the requested KubeMQ Server Address.

        :param str kubemq_address: KubeMQ server address. if None will be parsed from Config or environment parameter.
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """
        GrpcClient.__init__(self, encryptionHeader)
        if kubemq_address:
            self._kubemq_address = kubemq_address

    def send_event(self, event):
        # type: (Event) -> Result
        """Publish a single event using the KubeMQ."""
        try:
            result = self.get_kubemq_client().SendEvent(event.to_inner_event(), metadata=self._metadata)
            if result:
                return Result(inner_result=result)
        except Exception as e:
            logger.exception(
                "Sender received 'Result': Error:'%s'" % (
                    e
                ))
            raise

    def stream_event(self, events_stream, response_handler=None):
        def callback():
            for event in events_stream:
                yield event.to_inner_event()

        try:
            responses = self.get_kubemq_client().SendEventsStream(callback(), metadata=self._metadata)
            for response in responses:
                if response_handler:
                    result = Result(response)
                    logger.debug(
                        "Sender received 'Result': EventID:'%s', Sent:'%s', Error:'%s'" % (
                            result.event_id,
                            result.sent,
                            result.error
                        ))
                    response_handler(result)
        except Exception as e:
            logger.exception(
                "Sender received 'Result': Error:'%s'" % (
                    e
                ))
            raise

    def ping(self):
        """ping check connection to the kubemq"""
        ping_result = self.get_kubemq_client().Ping(Empty())
        logger.debug("event sender KubeMQ address:%s ping result:%s'" % (self._kubemq_address, ping_result))
        return ping_result



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os

_path = None  # ServerAddress.
_key = None  # RegistrationKey.
_cert = None  # Certificate file.


def get_server_address():
    global _path

    if _path:
        return _path

    _path = get_from_environment_variable("KubeMQServerAddress")
    if _path:
        return _path
    else:
        _path = get_from_environment_variable("KUBEMQSERVERADDRESS")
        if _path:
            return _path

    return _path


def get_registration_key():
    global _key

    if _key:
        return _key

    _key = get_from_environment_variable("KubeMQRegistrationKey")
    if _key:
        return _key
    else:
        _key = get_from_environment_variable("KUBEMQREGISTRATIONKEY")
        if _key:
            return _key
        else:
            _key = ""

    return _key


def get_certificate_file():
    global _cert

    if _cert:
        return _cert

    _cert = get_from_environment_variable("KubeMQCertificateFile")
    if _cert:
        return _cert
    else:
        _cert = get_from_environment_variable("KUBEMQCERTIFICATEFILE")
        if _cert:
            return _cert
        else:
            _cert = ""

    return _cert


def get_from_environment_variable(key):
    return os.environ.get(key)



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from kubemq.events.lowlevel.event import Event
from kubemq.events.lowlevel.sender import Sender


class Channel:
    """Sender with a set of predefined parameters"""
    sender = None

    def __init__(self, params=None, channel_name=None, client_id=None, store=None, kubemq_address=None,
                 encryptionHeader=None):
        """
        Initializes a new instance of the Events.Channel class using params OR "Manual" Parameters.

        :param params: ChannelParameters to use instead of the other params
        :param channel_name: Represents The channel name to send to using the KubeMQ.
        :param client_id: Represents the sender ID that the messages will be send under.
        :param store: Represents if the events should be send to persistence.
        :param kubemq_address: The address the of the KubeMQ including the GRPC Port ,Example: "LocalHost:50000".
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """

        if params:
            self.channel_name = params.channel_name
            self.client_id = params.client_id
            self.store = params.store
            self.kubemq_address = params.kubemq_address
            self.encryptionHeader = params.encryptionHeader
        else:
            self.channel_name = channel_name
            self.client_id = client_id
            self.store = store
            self.kubemq_address = kubemq_address
            self.encryptionHeader = encryptionHeader

        if not self.channel_name:
            raise ValueError("channel_name parameter is mandatory")

        self.sender = Sender(self.kubemq_address, self.encryptionHeader)

    def __repr__(self):
        return "<Channel channel_name:%s client_id:%s store:%s kubemq_address:%s encryptionHeader:%s>" % (
            self.channel_name,
            self.client_id,
            self.store,
            self.kubemq_address,
            self.encryptionHeader
        )

    def send_event(self, event):
        """Sending a new Event using KubeMQ."""
        low_level_event = self.create_low_level_event(event)
        return self.sender.send_event(low_level_event)

    def stream_event(self, stream, handler=None):
        """Sending a new Event using KubeMQ."""

        def stream_handler():
            for event in stream:
                yield self.create_low_level_event(event)

        return self.sender.stream_event(stream_handler(), handler)

    def create_low_level_event(self, notification):
        return Event(
            channel=self.channel_name,
            client_id=self.client_id,
            store=self.store,
            event_id=notification.event_id,
            body=notification.body,
            metadata=notification.metadata,
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from kubemq.grpc import Subscribe
from kubemq.subscription.subscribe_type import SubscribeType


class SubscribeRequest:
    """Represents a set of parameters which the Subscriber uses to subscribe to the KubeMQ."""

    def __init__(self, subscribe_type=None, client_id=None, channel=None, events_store_type=None,
                 events_store_type_value=None, group=""):
        self.subscribe_type = subscribe_type
        """Represents the type of Subscriber operation."""

        self.client_id = client_id
        """Represents an identifier that will subscribe to kubeMQ under."""

        self.channel = channel
        """Represents the channel name that will subscribe to under kubeMQ."""

        self.events_store_type = events_store_type
        """Represents the type of subscription to persistence"""

        self.events_store_type_value = events_store_type_value
        """Represents the value of subscription to persistence queue."""

        self.group = group
        """Represents the group the channel is assign to , if not filled will be empty string(no group)."""

    def from_inner_subscribe_request(self, inner):
        self.subscribe_type = SubscribeType(inner.SubscribeTypeData)
        self.client_id = inner.ClientID
        self.channel = inner.Channel
        self.group = inner.Group or ""
        self.events_store_type_value = inner.EventsStoreTypeValue

    def to_inner_subscribe_request(self):
        request = Subscribe()
        request.SubscribeTypeData = self.subscribe_type.value
        request.ClientID = self.client_id
        request.Channel = self.channel
        request.Group = self.group or ""
        request.EventsStoreTypeData = self.events_store_type.value
        request.EventsStoreTypeValue = self.events_store_type_value
        return request

    def is_valid_type(self, subscriber):
        if subscriber == "CommandQuery":
            return self.subscribe_type == SubscribeType.Commands or self.subscribe_type == SubscribeType.Queries
        else:  # subscriber == "Events"
            return self.subscribe_type == SubscribeType.Events or self.subscribe_type == SubscribeType.EventsStore



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from datetime import datetime

from kubemq.commandquery.request_receive import RequestReceive
from kubemq.grpc import Response as InnerResponse

epoch = datetime.utcfromtimestamp(0)


class Response:
    """The response that receiving from KubeMQ after sending a request."""

    def __init__(self, request):
        if isinstance(request, RequestReceive):
            self.request_id = request.request_id
            """Represents a Response identifier."""
            self.reply_channel = request.reply_channel
            """Channel name for the Response. Set and used internally by KubeMQ server."""

        elif isinstance(request, InnerResponse):

            self.client_id = request.ClientID or ""
            """Represents the sender ID that the Response will be send under."""

            self.request_id = request.RequestID
            """Represents a Response identifier."""

            self.reply_channel = request.ReplyChannel
            """Channel name for the Response. Set and used internally by KubeMQ server."""

            self.metadata = request.Metadata or ""
            """Represents text as str."""

            self.body = request.Body
            """Represents The content of the Response."""

            self.cache_hit = request.CacheHit
            """Represents if the response was received from Cache."""

            self.timestamp = datetime.fromtimestamp(request.Timestamp)
            """Represents if the response Time."""

            self.executed = request.Executed
            """Represents if the response was executed."""

            self.error = request.Error
            """Error message"""

            self.tags = request.Tags
            """Represents key value pairs that help distinguish the message"""

        else:
            raise Exception("Unknown type" + str(type(request)))

    def convert(self):
        return InnerResponse(
            ClientID=self.client_id or "",
            RequestID=self.request_id,
            ReplyChannel=self.reply_channel,
            Metadata=self.metadata or "",
            Body=self.body,
            CacheHit=self.cache_hit,
            Timestamp=int((self.timestamp - epoch).total_seconds()),
            Executed=self.executed,
            Error=self.error
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from kubemq.grpc import ReceiveMessage
from kubemq.grpc import StreamRequestTypeUnknown
from kubemq.grpc import RejectMessage
from kubemq.grpc import ModifyVisibility
from kubemq.grpc import AckMessage
from kubemq.grpc import QueueMessage
from kubemq.grpc import StreamQueueMessagesRequest
from kubemq.grpc import ResendMessage
from kubemq.grpc import SendModifiedMessage
from kubemq.queue.message import Message
from kubemq.tools.id_generator import get_next_id


class TransactionMessagesResponse:
    def __init__(self, stream_queue_messages_response=None, request_id=None, is_error=None, error_message=None,
                 message=None, stream_request=None):
        if stream_queue_messages_response:
            self.request_id = stream_queue_messages_response.RequestID
            """Represents Unique identifier for the Request."""

            self.is_error = stream_queue_messages_response.IsError
            """Returned from KubeMQ, false if no error."""

            self.error = stream_queue_messages_response.Error
            """Error message, valid only if IsError true."""

            self.message = stream_queue_messages_response.Message
            """"The received Message."""
            self.stream_request_type = stream_queue_messages_response.StreamRequestTypeData
            """Request action: ReceiveMessage, AckMessage, RejectMessage, ModifyVisibility, ResendMessage,  
            SendModifiedMessage, Unknown. """
        else:
            self.message = message
            self.is_error = is_error
            self.error = error_message
            self.request_id = request_id
            self.stream_request_type = stream_request

    def __repr__(self):
        return "<TransactionMessagesResponse request_id:%s is_error:%s error:%s message:%s stream_request_type:%s>" % (
            self.request_id,
            self.is_error,
            self.error,
            self.message,
            self.stream_request_type
        )


def create_stream_queue_message_receive_request(queue, visibility_seconds, wait_time_seconds=None):
    """Create StreamQueueMessageRequest for receive"""
    return StreamQueueMessagesRequest(
        ClientID=queue.client_id,
        Channel=queue.queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=ReceiveMessage,
        VisibilitySeconds=visibility_seconds,
        WaitTimeSeconds=wait_time_seconds or queue.WaitTimeSecondsQueueMessages,
        ModifiedMessage=QueueMessage(),
        RefSequence=0
    )


def create_stream_queue_message_check_call_is_in_transaction_request(_queue):
    """Create empty for checking check_call_is_in_transaction"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel=_queue.queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=StreamRequestTypeUnknown,
        VisibilitySeconds=20,
        WaitTimeSeconds=8,
        ModifiedMessage=QueueMessage(),
        RefSequence=0
    )


def create_stream_queue_message_ack_request(_queue, msg_sequence):
    """Create StreamQueueMessageRequest for ack"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel=_queue.queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=AckMessage,
        VisibilitySeconds=0,
        WaitTimeSeconds=0,
        ModifiedMessage=None,
        RefSequence=msg_sequence
    )


def create_stream_queue_message_reject_request(_queue, msg_sequence):
    """Create StreamQueueMessageRequest for reject"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel=_queue.queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=RejectMessage,
        VisibilitySeconds=0,
        WaitTimeSeconds=0,
        ModifiedMessage=None,
        RefSequence=msg_sequence
    )


def create_stream_queue_message_extend_visibility_request(_queue, visibility):
    """Create StreamQueueMessageRequest for extend_visibility"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel=_queue.queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=ModifyVisibility,
        VisibilitySeconds=visibility,
        WaitTimeSeconds=0,
        ModifiedMessage=None,
        RefSequence=0
    )


def create_stream_queue_message_resend_request(_queue, queue_name):
    """Create StreamQueueMessageRequest for resend"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel=queue_name,
        RequestID=get_next_id(),
        StreamRequestTypeData=ResendMessage,
        VisibilitySeconds=0,
        WaitTimeSeconds=0,
        ModifiedMessage=None,
        RefSequence=0
    )


def create_stream_queue_message_modify_request(_queue, message):
    """Create StreamQueueMessageRequest for modify"""
    return StreamQueueMessagesRequest(
        ClientID=_queue.client_id,
        Channel="",
        RequestID=get_next_id(),
        StreamRequestTypeData=SendModifiedMessage,
        VisibilitySeconds=0,
        WaitTimeSeconds=0,
        ModifiedMessage=message,
        RefSequence=0
    )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import threading
from kubemq.grpc import SendQueueMessageResult as SendQueueMessageResult
from kubemq.tools.id_generator import get_next_id as get_next_id


class SendMessageResult:
    def __init__(self, send_message_result=None):
        if send_message_result:
            self.message_id = send_message_result.MessageID
            """Represents Unique identifier for the Request."""

            self.expiration_at = send_message_result.ExpirationAt
            """Represents when the message is expired"""

            self.is_error = send_message_result.IsError
            """Returned from KubeMQ, false if no error."""

            self.sent_at = send_message_result.SentAt
            """Represents when the message was sent to kubemq."""

            self.delayed_to = send_message_result.DelayedTo
            """Represents if the message was delayed."""

            self.error = send_message_result.Error
            """Error message, valid only if IsError true."""

    def __repr__(self):
        return "<Message message_id:%s expiration_at:%s is_error:%s sent_at:%s delayed_to:%s error:%s>" % (
            self.message_id,
            self.expiration_at,
            self.is_error,
            self.sent_at,
            self.delayed_to,
            self.error,
        )

    def convert_to_send_queue_message_result(self):
        """Convert a SendMessageResult to an SendQueueMessageResult"""
        return SendQueueMessageResult(
            MessageID=self.message_id or get_next_id(),
            ExpirationAt=self.expiration_at.value or "",
            IsError=self.is_error or "",
            SentAt=self.sent_at,
            DelayedTo=self.delayed_to,
            Error=self.error
        )


def convert_from_send_queue_message_result(notification):
    """Convert SendQueueMessageResult to SendMessageResult"""
    result = SendMessageResult()
    result.message_id = notification.MessageID
    result.expiration_at = notification.ExpirationAt
    result.is_error = notification.IsError
    result.sent_at = notification.SentAt,
    result.delayed_to = notification.DelayedTo,
    result.error = notification.Error,
    return result



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class Request:
    """Represents the Request used in a Channel."""

    def __init__(self, request_id=None, metadata=None, body=None, tags=None):
        """
        Initializes a new instance of a Request for a Channel

        :param str request_id: Represents a Request identifier.
        :param str metadata: Represents metadata for a Request.
        :param bytes body: Represents The content of the Request.
        """
        self.request_id = request_id
        """Represents a Request identifier."""

        self.metadata = metadata
        """Represents metadata for a Request."""

        self.body = body
        """Represents The content of the Request."""

        self.tags = tags
        """Represents key value pairs that help distinguish the message"""

    def __repr__(self):
        return "<requst request_id:%s metadata:%s body:%s tags:%s>" % (
            self.request_id,
            self.metadata,
            self.body,
            self.tags
        )



import pathlib
from setuptools import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()



setuptools.setup(name='kubemq',
      version='2.3.0',
      description='KubeMQ SDK for Python',
	  long_description=README,
	  long_description_content_type="text/markdown",
      url='https://github.com/kubemq-io/kubemq-Python',
      author='KubeMQ',
      author_email='info@kubemq.io',
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=[
          'grpcio>=1.51.1',
          'protobuf>=4.21.0',
          'setuptools>=40.8.0',
          'PyJWT>=2.6.0',
      ],
      zip_safe=False)



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
import uuid
from kubemq.basic.grpc_client import GrpcClient
from kubemq.queue.send_message_result import SendMessageResult
from kubemq.queue.ack_all_messages_response import AckAllMessagesResponse
from kubemq.queue.message import to_queue_messages as to_queue_messages
from kubemq.queue.message import convert_queue_message_batch_request as convert_queue_message_batch_request
from kubemq.queue.receive_messages_response import ReceiveMessagesResponse as ReceiveMessagesResponse
from kubemq.grpc import Empty
from kubemq.grpc import ReceiveQueueMessagesRequest
from kubemq.grpc import AckAllQueueMessagesRequest
from kubemq.queue.send_batch_message_result import \
    convert_from_queue_messages_batch_response as convert_from_queue_messages_batch_response
from kubemq.tools.id_generator import get_next_id as get_next_id
from kubemq.queue.transaction import Transaction

logger = logging.getLogger(__name__)

class MessageQueue(GrpcClient):
    """Represents a MessageQueue pattern."""

    def __init__(self, queue_name=None, client_id=None, kubemq_address=None, max_number_of_messages=32,
                 wait_time_seconds_queue_messages=1, encryptionHeader=None):

        """
        Initializes a new MessageQueue using params .
        :param kubemq_address: The address the of the KubeMQ including the GRPC Port ,Example: "LocalHost:50000".
        :param client_id: Represents the sender ID that the messages will be send under.
        :param max_number_of_messages: Number of received messages.
        :param wait_time_seconds_queue_messages: Wait time for received messages.
        :param queue_name: Represents The queue name to send to using the KubeMQ.
        :param byte[] encryptionHeader: the encrypted header requested by kubemq authentication.
        """
        GrpcClient.__init__(self, encryptionHeader)
        if kubemq_address:
            self._kubemq_address = kubemq_address
        if client_id:
            self.client_id = client_id

        self.max_number_of_messages = max_number_of_messages

        self.wait_time_seconds_queue_messages = wait_time_seconds_queue_messages
        self.encryptionHeader = encryptionHeader
        self.queue_name = queue_name
        self.transaction = None

    def create_transaction(self):
        if self.transaction is None:
            self.transaction = Transaction(self)
        return self.transaction

    def send_queue_message(self, message):
        """Publish a single message using the KubeMQ."""
        try:
            message.queue = self.queue_name
            message.client_id = self.client_id
            inner_queue_message = message.convert_to_queue_message(self)
            inner_response = self.get_kubemq_client().SendQueueMessage(inner_queue_message, metadata=self._metadata)
            return SendMessageResult(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_queue_message_result:'%s'" % (e))
            raise

    def send_queue_messages_batch(self, messages):
        """"Publish a group of messages to a queue"""
        try:
            id = get_next_id()
            inner_response = self.get_kubemq_client().SendQueueMessagesBatch(
                convert_queue_message_batch_request(id, to_queue_messages(messages, self)), metadata=self._metadata)
            return convert_from_queue_messages_batch_response(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_queue_messages_batch:'%s'" % (e))
            raise

    def receive_queue_messages(self, number_of_messages=None):
        """Receive messages from queue"""
        try:
            id = get_next_id()
            inner_response = self.get_kubemq_client().ReceiveQueueMessages(
                self.convert_to_receive_queue_messages_request(id, False, number_of_messages), metadata=self._metadata)
            return ReceiveMessagesResponse(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_queue_messages:'%s'" % (e))
            raise

    def peek_queue_message(self, number_of_messages=None):
        """peek queue messages from queue"""
        try:
            id = get_next_id()
            inner_response = self.get_kubemq_client().ReceiveQueueMessages(
                self.convert_to_receive_queue_messages_request(id, True, number_of_messages), metadata=self._metadata)
            return ReceiveMessagesResponse(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_queue_messages_batch:'%s'" % (e))
            raise

    def ack_all_queue_messages(self):
        """acknowledge queue messages from queue"""
        try:
            inner_response = self.get_kubemq_client().AckAllQueueMessages(
                self.convert_to_ack_all_queue_message_request(), metadata=self._metadata)
            return AckAllMessagesResponse(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in ack_all_queue_messages:'%s'" % (e))
            raise

    def ping(self):
        """ping check connection to the kubemq"""
        ping_result = self.get_kubemq_client().Ping(Empty())
        logger.debug("MessageQueue KubeMQ address:%s ping result:%s'" % (self._kubemq_address, ping_result))
        return ping_result

    def convert_to_ack_all_queue_message_request(self):
        """create AckAllQueueMessagesRequest from MessageQueue"""
        return AckAllQueueMessagesRequest(
            RequestID=get_next_id(),
            ClientID=self.client_id,
            Channel=self.queue_name,
            WaitTimeSeconds=self.wait_time_seconds_queue_messages,
        )

    def convert_to_receive_queue_messages_request(self, request_id, is_peek=False, max_number_of_message=None):
        """create ReceiveQueueMessagesRequest from MessageQueue"""
        if max_number_of_message is None:
            max_number_of_message = self.max_number_of_messages
        return ReceiveQueueMessagesRequest(
            RequestID=request_id,
            ClientID=self.client_id,
            Channel=self.queue_name,
            MaxNumberOfMessages=max_number_of_message,
            WaitTimeSeconds=self.wait_time_seconds_queue_messages,
            IsPeak=is_peek,
        )

    def __repr__(self):
        return "<MessageQueue kubemq_address:%s client_id:%s max_number_of_messages:%s wait_time_seconds_queue_messages:%s>" % (
            self._kubemq_address,
            self.client_id,
            self.max_number_of_messages,
            self.wait_time_seconds_queue_messages
        )



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class RequestReceive:

    def __init__(self, inner_request):
        """
        :param inner_request: inner KubeMQ.Grpc.Request sent from KubeMQ.
        """

        self.request_id = inner_request.RequestID
        """Represents a Request identifier."""

        self.request_type = inner_request.RequestTypeData
        """Represents the type of request operation"""

        self.client_id = inner_request.ClientID
        """Represents the sender ID that the Request return from."""

        self.channel = inner_request.Channel
        """Represents The channel name to send to using the KubeMQ."""

        self.metadata = inner_request.Metadata
        """Represents text as str"""

        self.body = inner_request.Body
        """Represents The content of the KubeMQRequestReceive."""

        self.reply_channel = inner_request.ReplyChannel
        """Represents The channel name that the response returned from."""

        self.timeout = inner_request.Timeout
        """Represents the limit for waiting for response (Milliseconds)."""

        self.cache_key = inner_request.CacheKey
        """Represents if the request should be saved from Cache and under what "Key"(str) to save it."""

        self.cache_ttl = inner_request.CacheTTL
        """Cache time to live : for how long does the request should be saved in Cache."""

        self.tags = inner_request.Tags
        """Represents key value pairs that help distinguish the message"""



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class RequestParameters:
    def __init__(self, timeout=None, cache_key=None, cache_ttl=None):
        """
        Initializes a new instance of RequestParameters
        :param int timeout: Represents the limit for waiting for response (Milliseconds).
        :param str cache_key: Represents if the request should be saved from Cache and under what "Key"(str) to save it.
        :param int cache_ttl: Cache time to live : for how long does the request should be saved in Cache.
        """
        self.timeout = timeout
        """Represents the limit for waiting for response (Milliseconds)."""

        self.cache_key = cache_key
        """Represents if the request should be saved from Cache and under what "Key"(str) to save it."""

        self.cache_ttl = cache_ttl
        """Cache time to live : for how long does the request should be saved in Cache."""



# MIT License
#
# Copyright (c) 2018 KubeMQ
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from kubemq.grpc import Empty
from kubemq.basic.grpc_client import GrpcClient
from kubemq.commandquery.response import Response

logger = logging.getLogger(__name__)


class Initiator(GrpcClient):
    """Represents the instance that is responsible to send requests to the kubemq."""

    def __init__(self, kubemq_address=None, encryptionHeader=None):
        """
        Initialize a new Initiator.

        :param str kubemq_address: KubeMQ server address. if None will be parsed from Config or environment parameter.
        """
        GrpcClient.__init__(self, encryptionHeader)
        if kubemq_address:
            self._kubemq_address = kubemq_address

    def ping(self):
        """ping check connection to the kubemq"""
        ping_result = self.get_kubemq_client().Ping(Empty())
        logger.debug("Initiator KubeMQ address:%s ping result:%s'" % (self._kubemq_address, ping_result))
        return ping_result

    def send_request_async(self, request, handler):
        """Publish a single request using the KubeMQ."""

        def process_response(call):
            handler(Response(call.result()))

        try:
            inner_request = request.convert()
            call_future = self.get_kubemq_client().SendRequest.future(inner_request, None, self._metadata)
            call_future.add_done_callback(process_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_request_async'%s'" % (e))
            raise

    def send_request(self, request):
        """Publish a single request using the KubeMQ."""
        try:
            inner_request = request.convert()
            inner_response = self.get_kubemq_client().SendRequest(inner_request, None, self._metadata)
            return Response(inner_response)
        except Exception as e:
            logger.exception("Grpc Exception in send_request:'%s'" % (e))
            raise
